use axum::{
	extract::ws::{self,Message}, response, routing, Router
};
use futures_util::{stream::{SplitSink, SplitStream}, SinkExt, StreamExt};
use open_tavern::api::{Request, RequestError, Response};
use tokio::sync::{broadcast, mpsc};

#[tokio::main]
async fn main() {
	let (bsend, _) = broadcast::channel(10);
	let (msend, mrecv) = mpsc::channel(50);
	let bsend2 = bsend.clone();
	let app = Router::new()
		.route("/", routing::get(root))
		.route("/socket.js", routing::get(socket))
		.route("/ws", routing::get(move |w| ws_handler(w, msend, bsend2.clone().subscribe())));

	let listener = tokio::net::TcpListener::bind("0.0.0.0:3001").await.unwrap();
	let game = open_tavern::GameServer::new();
	tokio::spawn(game.server_loop(mrecv, bsend));

	axum::serve(listener, app).await.expect("axum server crashed, yaaaaay (unless i crashed him that yay)");
}

async fn ws_handler(ws: ws::WebSocketUpgrade, msend: mpsc::Sender<(String, Request)>, brecv: broadcast::Receiver<(Option<String>, Response)>) -> impl axum::response::IntoResponse {
	ws.on_upgrade(|w| handle_socket(w, msend, brecv))
}

async fn socket_receiver(mut recv: SplitStream<ws::WebSocket>, msend: mpsc::Sender<(String, Request)>, id: String) {
	while let Some(msg) = recv.next().await {
		if let Ok(msg) = msg {
			match msg {
				Message::Text(t) =>  {
					println!("Got message from {}: {}", &id, &t);
					let req = serde_json::from_str::<Request>(&t).unwrap_or_default();
					let erred = msend.send((id.clone(), req)).await.is_err();
					if erred {
						break;
					}
				}
				ws::Message::Binary(_) => todo!(),
				ws::Message::Ping(_) => todo!(),
				ws::Message::Pong(_) => todo!(),
				ws::Message::Close(_) => {
					// dont care if we fail the send as we are quitting regardless
					_ = msend.send((id.clone(), open_tavern::api::Request::Quit)).await;
					break;
				},
			}
		}
	}
}

async fn socket_sender(id: String, mut send: SplitSink<ws::WebSocket, ws::Message>, mut brecv: broadcast::Receiver<(Option<String>, Response)>) {
	while let Ok((to_id, msg)) = brecv.recv().await {
		if to_id.is_none() || to_id.map(|t| t == id).unwrap_or(false) {
			println!("Sending a message to {}: {:?}", &id, &msg);
			let err = send.send(ws::Message::Text(serde_json::to_string(&msg).unwrap())).await.is_err();
			if err || matches!(msg, Response::Shutdown) {
				_ = send.close().await;
				break;
			}
		}
	}
}

async fn handle_socket(mut socket: ws::WebSocket, msend: mpsc::Sender<(String, Request)>, brecv: broadcast::Receiver<(Option<String>, Response)>) {
	let mut id: Option<String> = None;
	loop {
		if let Some(msg) = socket.recv().await {
			if let Ok(msg) = msg {
				match msg {
					Message::Text(t) =>  {
						let req = serde_json::from_str::<Request>(&t).unwrap_or_default();
						println!("Got unauthorized message: {:?}", t);
						match req {
							// TODO: Actual signing in mechanism with multiple ids :)
							Request::Login(r) => if r.username == "rusty" || r.username == "honey" {
									_ = socket.send(Message::Text(
										serde_json::to_string(&Response::Login(open_tavern::api::login::LoginResult { success: true })).unwrap()
									)).await;
									id = Some(String::from(r.username));
									break;
								}
								else {
									_ = socket.send(Message::Text(
										serde_json::to_string(&Response::Login(open_tavern::api::login::LoginResult { success: false })).unwrap()
									)).await;
								},
							_ => {
								_ = socket.send(Message::Text(serde_json::to_string(&Response::Error(RequestError::InvalidRequest)).unwrap())).await;
							},
						}
					}
					ws::Message::Binary(_) => todo!(),
					ws::Message::Ping(_) => todo!(),
					ws::Message::Pong(_) => todo!(),
					ws::Message::Close(_) => break,
				};
			}
		}
		else {
			break;
		}
	}
	if let Some(id) = id {
		println!("Got id for socket: {}", &id);
		let (send, recv) = socket.split();
		tokio::spawn(socket_receiver(recv, msend, id.clone()));
		tokio::spawn(socket_sender(id, send, brecv));
	}
	println!("Done with so-cat");
}

async fn root() -> axum::response::Html<&'static str> {
	response::Html(include_str!("../assets/web/index.html"))
}

async fn socket() -> &'static str {
	include_str!("../assets/web/socket.js")
}

